/*
 Standalone Current Logger for ATtiny44A
 Version: 0.4
 Author: Alex from insideGadgets (www.insidegadgets.com)
 Created: 25/04/2012
 Last Modified: 28/07/2012
 
 This project uses an ATtiny44A with an OPA333 op-amp to log the current passing through a load.
 Logs 4096 readings in about 1 second when the button is pressed. 
 Plug in the USB, press the button and run the sclextract program to transfer the data to your computer.
  
 Remember to change your fuse bits so the ATtiny44A uses 16MHz crystal and divide clock by 8
 avrdude -p ATtiny44 -c usbtiny -U lfuse:w:0x7f:m -U hfuse:w:0xdf:m -U efuse:w:0xff:m
 
 */

// ATtiny24/44/84 Pin map
//
//               +-\/-+
//         VCC  1|o   |14  GND
//  XTAL   PB0  2|    |13  PA0 USB D+ 
//  XTAL   PB1  3|    |12  PA1 USB D-
//  Reset  PB3  4|    |11  PA2 Button
//  USB D+ PB2  5|    |10  PA3 ADC3
//  SPI CS PA7  6|    |9   PA4 SPI SCK
//  SPI DI PA6  7|    |8   PA5 SPI DO/LED
//               +----+
//

#define adcPin 3 // Analog input 3
#define buttonPin PA2 // Input/Output
#define usbminusPin PA1 // USB D-
#define spiCS PA7 // SPI Chip Select

#define F_CPU 16000000 // If this will be changed, modify it in usbconfig.h too

#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/sleep.h>
#include <math.h>
#include <util/delay.h>
#include "usbdrv/usbdrv.h"
#include "usbdrv/oddebug.h"
#include "setup.c"

volatile boolean f_pin = 0; // Pin change variable
int baseline = 13; // 13 * 0.001074 = ~14mV baseline
boolean inusbTransfer = true;
uint16_t readCount = 0;
uint16_t endCount = 4096;

// V-USB Transfer variables
#define USB_TRANSFER 1
static uchar replyBuf[16];
static uchar *nextDigit;
static uchar dataReceived = 0, dataLength = 0; // for USB_DATA_IN

// Convert a number to ASCII and add it to the replyBuf variable (it will only between 0 to 125)
void numberToascii(int number) {
	uchar diviser = 100;
	if ((number / diviser) != 0) {
		*++nextDigit = (number / diviser) + 48;
		*++nextDigit = ((number % diviser) / 10) + 48;
	}
	else if ((number / 10) != 0){
		*++nextDigit = (number / 10) + 48;
	}
	*++nextDigit = (number % 10) + 48;
}

// SPI Transfer
uint8_t spi_transfer(uint8_t data) {
	USIDR = data;
	USISR = _BV(USIOIF); // clear flag

	while ( (USISR & _BV(USIOIF)) == 0 ) {
		USICR = (1<<USIWM0)|(1<<USICS1)|(1<<USICLK)|(1<<USITC);
	}
	return USIDR;
}

// Send the next digit from the address
int sendnextDigit(uint16_t sramAddress) {
	PORTA &= ~(1<<spiCS); // CS Low, active
	spi_transfer(0x03);
	spi_transfer(sramAddress >> 8);
	spi_transfer(sramAddress & 0xFF);
	int memHighbyte = spi_transfer(0);
	int memLowbyte = spi_transfer(0);
	PORTA |= (1<<spiCS); // CS High, in-active
	
	int value = (memHighbyte << 8) | memLowbyte;
	
	// Minus the result from the baseline
	double calculateDouble = ((double) (value - baseline) * 0.01074);
	
	// Split the value into 2 parts
	int beforeDecimalpoint = (int) calculateDouble; // Before the decimal point
	int afterDecimalpoint = (int) ((calculateDouble - beforeDecimalpoint) * 1000); // After the decimal point
	
	// Fixes incorrect rounding after decimal point
	bool insertNumberzero = false;
	if (afterDecimalpoint < 100) {
		insertNumberzero = true;
	}
	
	// First part
	numberToascii(beforeDecimalpoint);
	
	if (afterDecimalpoint != 0) {
		*++nextDigit = 46; // Dot
		
		if (insertNumberzero == true) {
			*++nextDigit = 48; // 0
		}
		
		numberToascii(afterDecimalpoint);
	}
	
	*++nextDigit = '\r';
	*++nextDigit = '\n';
	*++nextDigit = '\0';
	
	return 0;
}

// This gets called when custom control message is received
USB_PUBLIC uchar usbFunctionSetup(uchar data[8]) {
	usbRequest_t *rq = (void *)data; // cast data to correct type
	
	if (rq->bRequest == USB_TRANSFER) { // custom command is in the bRequest field
		nextDigit = &replyBuf[-1];
		if (readCount < endCount) {
			int checkforZero = sendnextDigit(readCount);
			if (checkforZero == true) { // End if 0
				inusbTransfer = false;
			}
			readCount = readCount + 2;
		}
		else {
			inusbTransfer = false;
		}
		
		usbMsgPtr = replyBuf;
		
		return sizeof(replyBuf);
	}
	return 0;
}

// This gets called when data is sent from PC to the device
USB_PUBLIC uchar usbFunctionWrite(uchar *data, uchar len) {
	uchar i;
	
	for(i = 0; dataReceived < dataLength && i < len; i++, dataReceived++)
		replyBuf[dataReceived] = data[i];
	
	return (dataReceived == dataLength); // 1 if we received it all, 0 if not
}

// Transfers the EEPROM data to USB.
void transferData (void) {
	
	// Reset variables
	inusbTransfer = true;
	readCount = 0;

	while (inusbTransfer == true) {
		usbPoll();
	}
}

// Blink the LED x number of times with the pause calculated by the watchdog timer selected
void blinkLed (byte times, byte wdttime) {
	for (byte x = 0; x < times; x++) {
		DDRA |= (1<<PA5); // SPI DO output
		PORTA |= (1<<PA5); // LED on
		setup_watchdog(wdttime);
		system_sleep();
		
		PORTA &= ~(1<<PA5); // LED off
		DDRA &= ~(1<<PA5); // SPI DO input
		setup_watchdog(wdttime);
		system_sleep();
	}
	turnOffwatchdog();
	DDRA |= (1<<PA5); // SPI DO output
}

int main (void) {
	setup();
	
	DDRA &= ~(1<<usbminusPin); // Set Inputs
	DDRA &= ~(1<<buttonPin);
	
	// SPI setup
	DDRA |= (1<<PA5); // SPI DO output
	DDRA |= (1<<PA4); // SPI CLK output
	DDRA &= ~(1<<PA6); // SPI DI input
	PORTA |= (1<<PA6); // SPI DI input
	DDRA |= (1<<spiCS); // SPI CS output
	
	// Enable sequential mode by writing to the register
	PORTA &= ~(1<<spiCS); // CS Low, active
	spi_transfer(0x01);
	spi_transfer(0x40);
	PORTA |= (1<<spiCS); // CS High, in-active
	
	// Pin interrupt setup
	sbi(GIMSK,PCIE0); // Enable pin change interrupt
	sbi(PCMSK0,PCINT2); // Apply interrupt to button pin
	
	while(1) {
		DDRA &= ~(1<<PA5); // SPI DO input
		system_sleep();
		
		// Button was pressed, first test we are connected via USB by checking if the D- pin is high
		int usbConnected = PINA & (1<<usbminusPin);
		if (usbConnected) { // Send data from EEPROM to USB
			DDRA |= (1<<PA5); // SPI DO output
			
			// Increase clock speed to 16MHz
			CLKPR = (1<<CLKPCE); // Prescaler enable
			CLKPR = 0; // Clock division factor 1 (0000)
			
			// Transfer the data to the PC
			startUSB();
			transferData();
			
			// Decrease clock speed to 4MHz
			CLKPR = (1<<CLKPCE); // Prescaler enable
			CLKPR = (1<<CLKPS1); // Clock division factor 4 (0010)
		}
		else {
			f_pin = 0;
			setup_watchdog(T500MS);
			system_sleep();
			turnOffwatchdog();
			
			cbi(GIMSK,PCIE0); // Disable pin change interrupt
			DDRA |= (1<<PA5); // SPI DO output
			
			if (f_pin == 0) { // Button held down, record the baseline reading
				uint8_t adcLowbyte, adcHighbyte;
				
				// Turn on ADC
				sbi(ADCSRA, ADEN);
				sbi(ADCSRA, ADIE);
				
				// Start the ADC conversion and wait to be woken up by the ADC interrupt
				set_sleep_mode(SLEEP_MODE_ADC); // Set sleep mode ADC
				sleep_mode(); // System sleeps here
				
				// Read ADC bytes
				adcLowbyte = ADCL;
				adcHighbyte = ADCH;
				
				// Update the baseline
				baseline = (adcHighbyte << 8) | adcLowbyte;
				
				// Turn off ADC
				cbi(ADCSRA, ADSC);
				cbi(ADCSRA, ADEN);
				cbi(ADCSRA, ADIE);
				
				// Blink LED to confirm
				blinkLed(3, T128MS);
			}
			
			// Button let go, start logging
			else if (f_pin == 1) {
				int adcCount = 0;
				uint8_t adcLowbyte, adcHighbyte;
				
				// Turn on ADC
				sbi(ADCSRA, ADEN);
				sbi(ADCSRA, ADIE);
				
				// Start the ADC conversion and wait to be woken up by the ADC interrupt
				set_sleep_mode(SLEEP_MODE_ADC); // Set sleep mode ADC
				sleep_mode(); // System sleeps here
				
				// Log ADC 4,096 times for a 64Kbit SRAM (250uS for each ADC result)
				while (adcCount < 4096) {
					// Read ADC bytes
					adcLowbyte = ADCL;
					adcHighbyte = ADCH;
					
					// Write to SRAM
					PORTA &= ~(1<<spiCS); // CS Low, active
					spi_transfer(0x02);
					spi_transfer(adcCount >> 8);
					spi_transfer(adcCount & 0xFF);
					spi_transfer(adcHighbyte);
					spi_transfer(adcLowbyte);
					PORTA |= (1<<spiCS); // CS High, in-active
					
					adcCount = adcCount + 2;
					
					// Wait to be woken up by the ADC interrupt
					set_sleep_mode(SLEEP_MODE_ADC); // Set sleep mode ADC
					sleep_mode(); // System sleeps here
				}
				
				// Turn off ADC
				cbi(ADCSRA, ADSC);
				cbi(ADCSRA, ADEN);
				cbi(ADCSRA, ADIE);
			}
			
			sbi(GIMSK,PCIE0); // Enable pin change interrupt
		}
	}
	
	return 0;
}

ISR(WDT_vect) { }

// Pin change interrupt
ISR(PCINT0_vect) {
	f_pin = 1; // Set global button pressed (or released) flag
}

// ADC conversion complete interrupt
ISR(ADC_vect) { }
